import clang.cindex
from clang.cindex import AccessSpecifier, TypeKind


__primitive_types = set([
    TypeKind.VOID,
    TypeKind.BOOL,
    TypeKind.CHAR_U,
    TypeKind.UCHAR,
    TypeKind.CHAR16,
    TypeKind.CHAR32,
    TypeKind.USHORT,
    TypeKind.UINT,
    TypeKind.ULONG,
    TypeKind.ULONGLONG,
    TypeKind.UINT128,
    TypeKind.CHAR_S,
    TypeKind.SCHAR,
    TypeKind.WCHAR,
    TypeKind.SHORT,
    TypeKind.INT,
    TypeKind.LONG,
    TypeKind.LONGLONG,
    TypeKind.INT128,
    TypeKind.FLOAT,
    TypeKind.DOUBLE,
    TypeKind.LONGDOUBLE,
    TypeKind.NULLPTR,
])


def is_builtin(n):
    """ Test if a type is a simple types (integer, boolean, char, float)
    """
    return n.kind in __primitive_types


def is_definition(cursor):
    """Test if a cursor refers to a definition

    This occurs when the cursor has a definition, and shares the location of that definiton
    """
    defn = cursor.get_definition()
    return (defn is not None) and (cursor.location == defn.location)


def has_access(access):
    """ Test if a cursor has a access specifier
    """
    def _has_access(cursor):
        return cursor.access_specifier == access
    return _has_access


def is_kind(kind):
    """ Test if a cursor or type is of a particular kind
    """
    def _is_kind(cursor):
        return kind == cursor.kind
    return _is_kind


def has_storage_class(kind):
    """Check if the cursor has a particular (eg/ static) storage class
    """
    def _has_storage_class(cursor):
        return cursor.storage_class == kind
    return _has_storage_class


def is_decl(c):
    """Check if a cursor is a declaration
    """
    return c.kind.is_declaration()


def is_stmt(c):
    """Check if a cursor is a statement
    """
    return c.kind.is_declaration()



import clang.cindex
import os

__all__ = ['ClangDiagnosticException', 'parse_string', 'parse']


class ClangDiagnosticException(Exception):
    """Encapsulates Clang diagnostics as an exception
    """

    def __init__(self, diagnostic):
        self.diagnostic = diagnostic

    def __str__(self):
        s = ''
        for item in self.diagnostic:
            s += '%s\n' % item
        return s


def _ensure_parse_valid(tu):
    if len(tu.diagnostics) > 0:
        raise ClangDiagnosticException(tu.diagnostics)
    return tu


def parse_string(contents, name='tmp.cpp', **kwargs):
    """ Parse a string of C/C++ code
    """
    idx = clang.cindex.Index.create()
    tu = idx.parse(name, unsaved_files=[(name, contents)], **kwargs)
    return _ensure_parse_valid(tu)


def parse(name, **kwargs):
    """ Parse a C/C++ file
    """
    idx = clang.cindex.Index.create()
    assert os.path.exists(name)
    tu = idx.parse(name, **kwargs)
    return _ensure_parse_valid(tu)



def walk(predicate, cursor):
    """Yield all nodes found by recursively visiting the AST
    """
    return (c for c in cursor.walk_preorder() if predicate(c))


def iter_child_nodes(predicate, cursor):
    """Yield all direct child nodes of node
    """
    return (c for c in cursor.get_children() if predicate(c))



from clang.cindex import *
import re


__all__ = [
    'Matcher', 'UnlessMatcher', 'AnyOfMatcher', 'ChildAnyOfMatcher',
    'AnyBaseClassMatcher', 'NameMatcher', 'TypenameMatcher', 'AllOfTypeMatcher',
    'TypeTraversalMatcher', 'ReturnTypeTraversalMatcher', 'AnyParameterMatcher',
    'AncestorMatcher', 'TrueMatcher', 'LocationMatcher', 'ParentMatcher',
    'ParameterCountMatcher', 'CanonicalTypeTraversalMatcher',
    'PointeeTypeTraversalMatcher', 'ParameterMatcher'
]


class Matcher(object):
    """Base class for matchers
    """

    def __init__(self, *args):
        self.matchers = list(args)

    def __call__(self, cursor):
        assert(cursor is not None)
        ms = (m(cursor) for m in self.matchers)
        return all(ms)


class TrueMatcher(Matcher):
    """Matcher that always returns true
    """

    def __init__(self, *args):
        super(TrueMatcher, self).__init__()

    def __call__(self, cursor):
        return True


class UnlessMatcher(Matcher):
    """Inverts the match of the children
    """

    def __init__(self, *args):
        super(UnlessMatcher, self).__init__(*args)

    def __call__(self, cursor):
        return not super(UnlessMatcher, self).__call__(cursor)


class AnyOfMatcher(Matcher):
    def __init__(self, *args):
        super(AnyOfMatcher, self).__init__(*args)

    def __call__(self, cursor):
        assert(cursor is not None)
        return any(m(cursor) for m in self.matchers)


class ChildAnyOfMatcher(AnyOfMatcher):
    def __init__(self, *args):
        super(ChildAnyOfMatcher, self).__init__(*args)

    def children(self, cursor):
        return cursor.get_children()

    def __call__(self, cursor):
        assert(cursor is not None)
        assert(type(cursor) == Cursor)
        for c in self.children(cursor):
            if super(ChildAnyOfMatcher, self).__call__(c):
                return True
        return False


class AnyBaseClassMatcher(Matcher):
    def __init__(self, *args):
        super(AnyBaseClassMatcher, self).__init__(*args)

    def __call__(self, cursor):
        assert(cursor is not None)
        assert(type(cursor) == Cursor)
        for c in cursor.get_children():
            if c.kind == CursorKind.CXX_BASE_SPECIFIER:
                cdef = c.get_definition()
                if cdef is None:
                    continue
                if any(m(cdef) for m in self.matchers):
                    return True
                elif self(cdef):
                    return True
        return False


class NameMatcher(Matcher):
    def __init__(self, pattern):
        """Test if the  name refered to by the cursor matches a regex
        """
        super(NameMatcher, self).__init__()
        self.pattern = pattern

    def __call__(self, cursor):
        name = cursor.spelling or ''
        return re.match(self.pattern + '$', name) is not None


class TypenameMatcher(Matcher):
    def __init__(self, pattern):
        """Test if the typename refered to by the cursor matches a regex
        """
        super(TypenameMatcher, self).__init__([])
        self.pattern = pattern

    def __call__(self, cursor):
        typename = cursor.type.spelling or ''
        return re.match(self.pattern + '$', typename) is not None


class AllOfTypeMatcher(object):
    def __init__(self, *args):
        self.matchers = list(args)

    def __call__(self, t):
        assert(t is not None)
        assert(type(t) == Type)
        return all(m(t) for m in self.matchers)


class TypeTraversalMatcher(object):
    def __init__(self, matcher):
        self.matcher = matcher

    def traverse(self, cursor):
        return cursor.type

    def __call__(self, cursor):
        assert(cursor is not None)
        assert(type(cursor) == Cursor)
        return self.matcher(self.traverse(cursor))


class PointeeTypeTraversalMatcher(object):
    def __init__(self, inner):
        self.inner = inner

    def __call__(self, t):
        assert(t is not None)
        assert(type(t) == Type)
        return self.inner(t.get_pointee())


class ReturnTypeTraversalMatcher(TypeTraversalMatcher):
    def __init__(self, matcher):
        super(ReturnTypeTraversalMatcher, self).__init__(matcher)

    def traverse(self, cursor):
        return cursor.result_type


class AnyParameterMatcher(object):
    def __init__(self, matcher):
        self.matcher = matcher

    def __call__(self, cursor):
        assert(cursor is not None)
        assert(type(cursor) == Cursor)
        for a in cursor.get_arguments():
            if self.matcher(a):
                return True
        return False


class AncestorMatcher(Matcher):
    def __init__(self, m):
        super(AncestorMatcher, self).__init__(m)

    def __call__(self, cursor):
        assert(cursor is not None)
        assert(type(cursor) == Cursor)
        c = cursor
        while c is not None:
            if super(AncestorMatcher, self).__call__(c):
                return True
            c = c.semantic_parent
        return False


class ParentMatcher(Matcher):
    def __init__(self, m):
        super(ParentMatcher, self).__init__(m)

    def __call__(self, cursor):
        assert(cursor is not None)
        assert(type(cursor) == Cursor)
        p = cursor.semantic_parent
        if p is None:
            return False
        return super(ParentMatcher, self).__call__(p)


class LocationMatcher(Matcher):
    def __init__(self, pattern):
        super(LocationMatcher, self).__init__()
        self.pattern = pattern

    def __call__(self, cursor):
        try:
            fname = cursor.location.file.name
            return re.match(self.pattern, fname)
        except:
            return False


class ParameterCountMatcher(Matcher):
    def __init__(self, N):
        super(ParameterCountMatcher, self).__init__()
        self.N = N

    def __call__(self, cursor):
        assert(cursor is not None)
        return self.N == len(list(cursor.get_arguments()))


class CanonicalTypeTraversalMatcher(TypeTraversalMatcher):
    def __init__(self, matcher):
        super(CanonicalTypeTraversalMatcher, self).__init__(matcher)

    def traversal(self, t):
        return t.get_canonical()

    def __call__(self, t):
        return self.matcher(self.traversal(t))


class ParameterMatcher(Matcher):
    def __init__(self, N, inner):
        super(ParameterMatcher, self).__init__(inner)
        self.N = N

    def __call__(self, cursor):
        try:
            arg = list(cursor.get_arguments())[self.N]
            return super(ParameterMatcher, self).__call__(arg)
        except:
            return False



from clang.cindex import CursorKind
from .predicates import *
from .internal import *

__all__ = [
    'allOf', 'anyOf', 'hasType', 'anything', 'hasAnyParameter', 'builtinType',
    'classTemplateDecl', 'cxxRecordDecl', 'cxxConstructorDecl',
    'cxxDestructorDecl', 'cxxMethodDecl', 'decl', 'enumDecl', 'fieldDecl',
    'functionDecl', 'has', 'hasName', 'hasReturnType', 'hasStaticStorageDuration',
    'hasTypename', 'isDerivedFrom', 'isSameOrDerivedFrom', 'namespaceDecl',
    'recordDecl', 'stmt', 'typedefDecl', 'unless', 'isDefinition', 'hasAncestor',
    'isExpansionInFileMatching', 'varDecl', 'hasParent', 'parameterCountIs',
    'hasCanonicalType', 'isStruct', 'isClass', 'isPublic', 'isProtected', 'isPrivate',
    'pointerType', 'pointee', 'hasParameter'
]


def allOf(*args):
    """Matches if all of the argument matchers match

    >>> from glud import *
    >>> config = '''
    ...  class X;
    ...  class Y;
    ... '''
    >>> m = allOf(cxxRecordDecl(), hasName('X'))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return Matcher(*args)


def anyOf(*args):
    """Matches if any of the argument matchers match

    >>> from glud import *
    >>> config = '''
    ...  int x;
    ...  class Y {};
    ...  enum Z {};
    ... '''
    >>> m = anyOf(varDecl(), isClass())
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    x
    Y
    """
    return AnyOfMatcher(*args)


def anything():
    """Matches anything

    >>> from glud import *
    >>> config = '''
    ... namespace W {
    ...  int x;
    ...  class Y;
    ...  enum Z {};
    ... }
    ... '''
    >>> m = allOf(anything(), hasParent(namespaceDecl()))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    x
    Y
    Z
    """
    return TrueMatcher()


def builtinType(*args):
    """Matches builtin primitive types (eg/ integers, booleans and float)

    >>> from glud import *
    >>> config = '''
    ...  int x;
    ...  bool y;
    ... '''
    >>> m = varDecl(hasType(builtinType()))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    x
    y
    """
    return AllOfTypeMatcher(is_builtin, *args)


def classTemplateDecl(*args):
    """Match C++ template class declarations

    >>> from glud import *
    >>> config = '''
    ...  template<typename T> class X {};
    ...  class Y {};
    ... '''
    >>> m = classTemplateDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return Matcher(is_kind(CursorKind.CLASS_TEMPLATE), *args)


def cxxConstructorDecl(*args):
    """Match C++ constructors

    >>> from glud import *
    >>> config = '''
    ...  class X {
    ...    X();
    ...  };
    ...  class Y {};
    ... '''
    >>> m = cxxConstructorDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return Matcher(is_kind(CursorKind.CONSTRUCTOR), *args)


def cxxDestructorDecl(*args):
    """Match C++ destructors

    >>> from glud import *
    >>> config = '''
    ...  class X {
    ...    ~X();
    ...  };
    ...  class Y {};
    ... '''
    >>> m = cxxDestructorDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    ~X
    """
    return Matcher(is_kind(CursorKind.DESTRUCTOR), *args)


def cxxMethodDecl(*args):
    """Match C++ methods

    >>> from glud import *
    >>> config = '''
    ...  class X {
    ...    void u();
    ...    void v();
    ...  };
    ... '''
    >>> m = cxxMethodDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    u
    v
    """
    return Matcher(is_kind(CursorKind.CXX_METHOD), *args)


def cxxRecordDecl(*args):
    """Matches C++ class declarations.

    >>> from glud import *
    >>> config = '''
    ...  class W;
    ...  template<typename T> class X {};
    ...  struct Y {};
    ...  union Z {};
    ... '''
    >>> m = cxxRecordDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    W
    X
    """
    return Matcher(anyOf(
                is_kind(CursorKind.CLASS_DECL),
                is_kind(CursorKind.CLASS_TEMPLATE)), *args)


def decl(*args):
    """Match any declaration

    >>> from glud import *
    >>> config = '''
    ...  class X {};
    ...  struct Y {};
    ...  enum Z {};
    ... '''
    >>> m = decl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    Y
    Z
    """
    return Matcher(is_decl, *args)


def enumDecl(*args):
    """Match enumerations

    >>> from glud import *
    >>> config = '''
    ...  enum X {};
    ...  enum class Y {};
    ... '''
    >>> m = enumDecl()
    >>> for c in parse_string(config, args='-x c++ -std=c++11'.split()).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    Y
    """
    return Matcher(is_kind(CursorKind.ENUM_DECL), *args)


def fieldDecl(*args):
    """Match struct / class fields

    >>> from glud import *
    >>> config = '''
    ...  struct X {
    ...   int u;
    ...   int v;
    ...  };
    ... '''
    >>> m = fieldDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    u
    v
    """
    return Matcher(is_kind(CursorKind.FIELD_DECL), *args)


def functionDecl(*args):
    """Match function declarations

    >>> from glud import *
    >>> config = '''
    ...  int u();
    ...  int v();
    ... '''
    >>> m = functionDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    u
    v
    """
    return Matcher(is_kind(CursorKind.FUNCTION_DECL), *args)


def has(*args):
    """Match if a cursor has a child that matches

    >>> from glud import *
    >>> config = '''
    ...  class X {
    ...   void f();
    ...  };
    ...  class Y;
    ... '''
    >>> m = cxxRecordDecl(has(cxxMethodDecl()))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return ChildAnyOfMatcher(*args)


def hasAncestor(matcher):
    """Matches if the current cursor has an ancestor that matches

    >>> from glud import *
    >>> config = '''
    ... namespace X {
    ...   class Y {};
    ... }
    ... class Z {};
    ... '''
    >>> m = cxxRecordDecl(
    ...         hasName('Y'),
    ...         hasAncestor(namespaceDecl(hasName('X'))))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    Y
    """
    return AncestorMatcher(matcher)


def hasAnyParameter(matcher):
    """Match if any method or function argument matches

    >>> from glud import *
    >>> config = '''
    ... void f();
    ... void g(int);
    ... '''
    >>> m = functionDecl(hasAnyParameter(hasType(builtinType())))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    g
    """
    return AnyParameterMatcher(matcher)


def hasCanonicalType(m):
    """Matches if a cursor has the specified number of arguments

    >>> from glud import *
    >>> config = '''
    ...  namespace X {
    ...   struct Y;
    ...   Y f();
    ...  }
    ... '''
    >>> m = functionDecl(hasReturnType(hasCanonicalType(hasName('X::Y'))))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    f
    """
    return CanonicalTypeTraversalMatcher(m)


def hasName(name):
    """Match a cursors spelling against a pattern

    >>> from glud import *
    >>> config = '''
    ...  class X {};
    ...  class Y {};
    ... '''
    >>> m = cxxRecordDecl(hasName('X'))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return NameMatcher(name)


def hasParameter(N, inner):
    """Matches if the N-th parameter matches the inner matcher

    >>> from glud import *
    >>> config = '''
    ... void f(int x);
    ... void g(int y, int x);
    ... void h();
    ... '''
    >>> m = functionDecl(hasParameter(1, hasName('x')))
    >>> tuc = parse_string(config).cursor
    >>> for c in tuc.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    g
    """
    return ParameterMatcher(N, inner)


def hasParent(*args):
    """Matches if the direct parent node matches

    >>> from glud import *
    >>> config = '''
    ... namespace X {
    ...   int a;
    ... }
    ... int b;
    ... '''
    >>> m = varDecl(hasParent(namespaceDecl(hasName('X'))))
    >>> tuc = parse_string(config).cursor
    >>> for c in tuc.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    a
    """
    return ParentMatcher(*args)


def hasReturnType(matcher):
    """Match a function/method with a specified return type

    >>> from glud import *
    >>> config = '''
    ... class X {};
    ... X u();
    ... int v();
    ... '''
    >>> m = functionDecl(hasReturnType(builtinType()))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    v
    """
    return ReturnTypeTraversalMatcher(matcher)


def hasStaticStorageDuration():
    """Match an item has static storage duration

    >>> from glud import *
    >>> config = '''
    ... class X {
    ...  static void u();
    ...  void v();
    ... };
    ... '''
    >>> m = cxxMethodDecl(
    ...         hasStaticStorageDuration())
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    u
    """
    return Matcher(has_storage_class(clang.cindex.StorageClass.STATIC))


def hasType(matcher):
    """Matches if the type associated with the current cursor matches

    >>> from glud import *
    >>> config = '''
    ...  int x;
    ...  long y;
    ... '''
    >>> m = varDecl(hasType(hasName('int')))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    x
    """
    return TypeTraversalMatcher(matcher)


def hasTypename(typename):
    """Match if the spelling of the type of a cursor matches a pattern

    >>> from glud import *
    >>> config = '''
    ... namespace X {
    ...  class Y {};
    ... }
    ... class Y {};
    ... '''
    >>> m = cxxRecordDecl(hasTypename('X::Y'))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.type.spelling)
    X::Y
    """
    return TypenameMatcher(typename)


def isClass():
    """Matches if a cursor is a class

    >>> from glud import *
    >>> config = '''
    ...  class X;
    ...  struct Y;
    ... '''
    >>> m = isClass()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return Matcher(is_kind(CursorKind.CLASS_DECL))


def isDefinition():
    """Matches if the cursor is a definition

    >>> from glud import *
    >>> config = '''
    ... class X {};
    ... class Y;
    ... '''
    >>> m = cxxRecordDecl(isDefinition())
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return Matcher(is_definition)


def isDerivedFrom(name):
    """Match if a C++ type inherits from a named class

    >>> from glud import *
    >>> config = '''
    ... class X {};
    ... class Y : public X {};
    ... class Z : public Y {};
    ... '''
    >>> m = cxxRecordDecl(isDerivedFrom('X'))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    Y
    Z
    """
    return AnyBaseClassMatcher(hasTypename(name))


def isExpansionInFileMatching(pattern):
    """Matches if the nodes location matches a pattern

    >>> from glud import *
    >>> config = '''
    ... class X;
    ... '''
    >>> m = isExpansionInFileMatching('tmp.cpp')
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return LocationMatcher(pattern)


def isPrivate():
    """Test if a cursor is private

    >>> from glud import *
    >>> config = '''
    ...  class X {
    ...   private:
    ...    int y;
    ...  };
    ... '''
    >>> m = fieldDecl(isPrivate())
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    y
    """
    return Matcher(has_access(AccessSpecifier.PRIVATE))


def isProtected():
    """Test if a cursor is protected

    >>> from glud import *
    >>> config = '''
    ...  class X {
    ...   protected:
    ...    int y;
    ...  };
    ... '''
    >>> m = fieldDecl(isProtected())
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    y
    """
    return Matcher(has_access(AccessSpecifier.PROTECTED))


def isPublic():
    """Test if a cursor is public

    >>> from glud import *
    >>> config = '''
    ...  class X {
    ...   public:
    ...    int y;
    ...  };
    ... '''
    >>> m = fieldDecl(isPublic())
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    y
    """
    return Matcher(has_access(AccessSpecifier.PUBLIC))


def isSameOrDerivedFrom(name):
    """Match if derives-from (or is-a) class with a given name

    >>> from glud import *
    >>> config = '''
    ... class X {};
    ... class Y : public X {};
    ... class Z : public Y {};
    ... '''
    >>> m = cxxRecordDecl(isSameOrDerivedFrom('X'))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    Y
    Z
    """
    return anyOf(hasTypename(name), isDerivedFrom(name))


def isStruct():
    """Matches if a cursor is a struct

    >>> from glud import *
    >>> config = '''
    ...  class X;
    ...  struct Y;
    ... '''
    >>> m = isStruct()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    Y
    """
    return Matcher(is_kind(CursorKind.STRUCT_DECL))


def namespaceDecl(*args):
    """Match a C++ namespace declaration

    >>> from glud import *
    >>> config = '''
    ... namespace X { }
    ... '''
    >>> m = namespaceDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return Matcher(is_kind(CursorKind.NAMESPACE), *args)


def parameterCountIs(N):
    """Matches if a cursor has the specified number of arguments

    >>> from glud import *
    >>> config = '''
    ...  int f();
    ...  int g(int);
    ...  int h(int, int);
    ... '''
    >>> m = functionDecl(parameterCountIs(1))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    g
    """
    return ParameterCountMatcher(N)


def pointee(inner):
    """Traverse from a pointer to the dereferenced type

    >>> from glud import *
    >>> config = '''
    ...  int *x;
    ...  void *y;
    ... '''
    >>> m = varDecl(hasType(pointerType(pointee(hasName('int')))))
    >>> for c in parse_string(config, args='-x c++ -std=c++11'.split()).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    x
    """
    return PointeeTypeTraversalMatcher(inner)


def pointerType(*args):
    """Test if a cursor has pointer type

    >>> from glud import *
    >>> config = '''
    ...  int w;
    ...  int* x;
    ...  int** y;
    ...  int& z = w;
    ... '''
    >>> m = varDecl(hasType(pointerType()))
    >>> for c in parse_string(config, args='-x c++ -std=c++11'.split()).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    x
    y
    """
    return AllOfTypeMatcher(is_kind(TypeKind.POINTER), *args)


def recordDecl(*args):
    """Matches class, struct, and union declarations.

    >>> from glud import *
    >>> config = '''
    ...  class W;
    ...  template<typename T> class X {};
    ...  struct Y {};
    ...  union Z {};
    ... '''
    >>> m = recordDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    W
    X
    Y
    Z
    """
    return Matcher(anyOf(
                is_kind(CursorKind.STRUCT_DECL),
                is_kind(CursorKind.UNION_DECL),
                is_kind(CursorKind.CLASS_DECL),
                is_kind(CursorKind.CLASS_TEMPLATE)), *args)


def stmt(*args):
    """Matches statements

    >>> from glud import *
    >>> config = '''
    ... void f() { }
    ... '''
    >>> m = stmt()
    >>> i = 0
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         i += 1
    >>> print(i)
    1
    """
    return Matcher(is_stmt, *args)


def typedefDecl(*args):
    """Matches typedef declarations

    >>> from glud import *
    >>> config = '''
    ... typedef int X;
    ... '''
    >>> m = typedefDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    """
    return Matcher(is_kind(CursorKind.TYPEDEF_DECL), *args)


def unless(*args):
    """Inverts the match of the children

    >>> from glud import *
    >>> config = '''
    ... class X { };
    ... class Y {};
    ... class Z {};
    ... '''
    >>> m = cxxRecordDecl(unless(hasName('Y')))
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    X
    Z
    """
    return UnlessMatcher(*args)


def varDecl(*args):
    """Matches variable declarations

    >>> from glud import *
    >>> config = '''
    ...  int a;
    ... '''
    >>> m = varDecl()
    >>> for c in parse_string(config).cursor.walk_preorder():
    ...     if m(c):
    ...         print(c.spelling)
    a
    """
    return Matcher(is_kind(CursorKind.VAR_DECL), *args)



from asciitree import draw_tree


def dump(cursor):
    """ Display the AST represented by the cursor
    """

    def node_children(node):
        return list(node.get_children())

    def print_node(node):
        text = node.spelling or node.displayname
        kind = str(node.kind).split('.')[1]
        return '{} {}'.format(kind, text)

    return draw_tree(cursor, node_children, print_node)



from setuptools import setup, find_packages
import re
import io
import os
import sys


def read_version():
    with io.open('./glud/version.py', encoding='utf8') as version_file:
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file.read(), re.M)
        if version_match:
            version = version_match.group(1)
        else:
            raise runtimeerror("unable to find version string.")
    return version


def read(filename):
    path = os.path.join(os.path.dirname(__file__), filename)
    with io.open(path, encoding='utf8') as fh:
        return fh.read()


def requires():
    req = ['asciitree']
    if sys.version_info.major == 2:
        return req + ['clang']
    else:
        return req + ['libclang-py3']


setup(
    name="glud",
    install_requires=requires(),
    description="Functional tools for matching nodes in the clang AST",
    long_description=read('README.rst'),
    version=read_version(),
    author="Andrew Walker",
    author_email="walker.ab@gmail.com",
    url="http://github.com/AndrewWalker/glud",
    license="MIT",
    keywords=['libclang', 'clang', 'AST'],
    packages=find_packages(exclude=['tests']),
    classifiers=[
        'Topic :: Software Development :: Compilers',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Operating System :: POSIX :: Linux',
    ],
    test_suite='tests',
    zip_safe=False
)
